/*************************************************
* Disk Storage Allocator Source File             *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/mmap_mem.h>
#include <opencl/exceptn.h>
#include <opencl/util.h>
#include <cstring>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>

namespace OpenCL {

/*************************************************
* Constructor                                    *
*************************************************/
MmapDisk_Allocator::MmapDisk_Allocator(const std::string& dir)
   : ManagedAllocator(true, getpagesize())
   {
   if(getenv("TMPDIR"))
      path = getenv("TMPDIR");
   else if(dir != "")
      path = dir;
   else
      path = "/tmp";
   path += "/opencl_XXXXXX";
   lock = get_mutex();

   cache.resize(CACHE_SIZE);
   for(u32bit j = 0; j != cache.size(); j++)
      {
      cache[j].length = getpagesize();
      cache[j].buf = mmap_block(cache[j].length);
      cache[j].fd = allocated[cache[j].buf].first;
      allocated.erase(allocated.find(cache[j].buf));
      }
   }

/*************************************************
* Destructor                                     *
*************************************************/
MmapDisk_Allocator::~MmapDisk_Allocator()
   {
   for(u32bit j = 0; j != cache.size(); j++)
      munmap_block(cache[j].buf, cache[j].length);
   }

/*************************************************
* Guess the page size of the system              *
*************************************************/
u32bit MmapDisk_Allocator::getpagesize() const
   {
   static const u32bit MINIMUM_PAGE_SIZE = 512;
   u32bit pagesize = sysconf(_SC_PAGE_SIZE);
   if(pagesize >= MINIMUM_PAGE_SIZE && power_of_2(pagesize))
      return pagesize;
   return 4096;
   }

/*************************************************
* Allocation                                     *
*************************************************/
void* MmapDisk_Allocator::alloc_block(u32bit n) const
   {
   lock->lock();
   for(u32bit j = 0; j != cache.size(); j++)
      {
      if(cache[j].length == n && cache[j].buf != 0)
         {
         void* ptr = cache[j].buf;
         allocated[ptr] = std::pair<int, u32bit>(cache[j].fd, cache[j].length);
         cache[j].buf = 0;
         cache[j].length = 0;
         cache[j].fd = 0;
         lock->unlock();
         return ptr;
         }
      }
   lock->unlock();

   return mmap_block(n);
   }

/*************************************************
* Deallocation                                   *
*************************************************/
void MmapDisk_Allocator::dealloc_block(void* ptr, u32bit n) const
   {
   if(ptr == 0) return;

   lock->lock();

   if(allocated.find(ptr) == allocated.end())
      throw Exception("MmapDisk_Allocator: pointer not found");
   if(allocated[ptr].second != n)
      throw Exception("MmapDisk_Allocator: bad length for memory block");

   for(u32bit j = 0; j != cache.size(); j++)
      if(cache[j].buf == 0)
         {
         cache[j].buf = ptr;
         cache[j].length = n;
         cache[j].fd = allocated[ptr].first;
         lock->unlock();
         return;
         }
   lock->unlock();

   munmap_block(ptr, n);
   }

/*************************************************
* Memory Map a File into Memory                  *
*************************************************/
void* MmapDisk_Allocator::mmap_block(u32bit n) const
   {
   char* filepath = new char[path.length() + 1];
   std::strcpy(filepath, path.c_str());

   mode_t old_umask = umask(077);
   int file = mkstemp(filepath);
   if(file == -1)
      throw Exception("MmapDisk_Allocator: Could not create file");

   if(unlink(filepath))
      throw Exception("MmapDisk_Allocator: Could not unlink file");
   delete[] filepath;
   umask(old_umask);

   lseek(file, n, SEEK_SET);
   if(write(file, "\0", 1) != 1)
      throw Exception("MmapDisk_Allocator: Could not write to file");
   void* ptr = mmap(0, n, PROT_READ | PROT_WRITE, MAP_SHARED, file, 0);
   if(ptr == (void*)MAP_FAILED)
      throw Exception("MmapDisk_Allocator: Could not map file");

   lock->lock();
   allocated[ptr] = std::make_pair<int, u32bit>(file, n);
   lock->unlock();

   return ptr;
   }

/*************************************************
* Remove a Memory Map                            *
*************************************************/
void MmapDisk_Allocator::munmap_block(void* ptr, u32bit n) const
   {
   static const u32bit OVERWRITE_PASSES = 7;
   static const byte PATTERNS[12] = { 0x00, 0xFF, 0xAA, 0x55, 0x11, 0xEE, 
                                      0xDE, 0xAD, 0x77, 0x88, 0x33, 0xCC };

   if(ptr == 0) return;

   for(u32bit j = 0; j != OVERWRITE_PASSES; j++)
      {
      std::memset(ptr, PATTERNS[j % 12], n);
      if(msync(ptr, n, MS_SYNC))
         throw Exception("MmapDisk_Allocator: Sync operation failed");
      }
   std::memset(ptr, 0, n);
   if(msync(ptr, n, MS_SYNC))
      throw Exception("MmapDisk_Allocator: Sync operation failed");

   if(munmap(ptr, n))
      throw Exception("MmapDisk_Allocator: Could not unmap file");

   lock->lock();
   if(close(allocated[ptr].first))
      {
      lock->unlock();
      throw Exception("MmapDisk_Allocator: Could not close file");
      }
   lock->unlock();
   }

}
