/*************************************************
* Unix EntropySource Source File                 *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#define _POSIX_C_SOURCE 199506
#include <opencl/unix_rnd.h>
#include <opencl/util.h>
#include <stdio.h>

namespace OpenCL {

namespace {

bool Source_Cmp(const Source& a, const Source& b)
   { return (a.priority < b.priority); }

}

/*************************************************
* Sources List                                   *
*************************************************/
const Source Unix_EntropySource::DEFAULT_SOURCES[] = {
   { "vmstat",      "",                  1 },
   { "uptime",      "",                  1 },
   { "netstat",     "-in",               1 },
   { "netstat",     "-s",                2 },
   { "ipcs",        "-a",                2 },
   { "ps",          "-A",                2 },
   { "uname",       "-a",                2 },
   { "ps",          "aux",               3 },
   { "rpcinfo",     "-p localhost",      3 },
   { "dmesg",       "",                  3 },
   { "finger",      "",                  3 },
   { "who",         "",                  3 },
   { "arp",         "",                  3 },
   { "last",        "",                  4 },
   { "df",          "",                  4 },
   { "lsof",        "",                  5 }
};

/*************************************************
* Unix_EntropySource Constructor                 *
*************************************************/
Unix_EntropySource::Unix_EntropySource(bool add_defaults)
   {
   if(add_defaults)
      add_sources(DEFAULT_SOURCES, sizeof(DEFAULT_SOURCES) / 
                  sizeof(DEFAULT_SOURCES[0]));
   position = 0;
   }

/*************************************************
* Unix_EntropySource Constructor                 *
*************************************************/
Unix_EntropySource::Unix_EntropySource(const Source srcs[], u32bit count,
                                       bool add_defaults)
   {
   add_sources(srcs, count);
   if(add_defaults)
      add_sources(DEFAULT_SOURCES, sizeof(DEFAULT_SOURCES) / sizeof(Source));
   }

/*************************************************
* Fast Poll                                      *
*************************************************/
u32bit Unix_EntropySource::fast_poll(byte out[], u32bit length)
   {
   static const u32bit MAX_PRIORITY = 2;
   return gather(out, length, MAX_PRIORITY);
   }

/*************************************************
* Slow Poll                                      *
*************************************************/
u32bit Unix_EntropySource::slow_poll(byte out[], u32bit length)
   {
   static const u32bit MAX_PRIORITY = 4;
   return gather(out, length, MAX_PRIORITY);
   }

/*************************************************
* Slow Poll                                      *
*************************************************/
u32bit Unix_EntropySource::gather(byte out[], u32bit length, u32bit prio)
   {
   u32bit got = 0;
   for(u32bit j = 0; j != sources.size(); j++)
      {
      if((sources[j].priority > prio && got >= length) ||
         (sources[j].priority > prio + 2))
         break;
      got += gather_entropy(sources[j]);
      }
   u32bit copied = std::min(BUFFERSIZE - position, length);
   copy_mem(out, buffer + position, copied);
   position = (position + copied) % BUFFERSIZE;
   return copied;
   }

/*************************************************
* Gather Entropy from a Source                   *
*************************************************/
u32bit Unix_EntropySource::gather_entropy(const Source& src)
   {
   std::string command = src.name + " " + src.args + " 2>&1";

   SecureBuffer<byte, BUFFERSIZE> output;
   u32bit total_output = 0;
   FILE* pipe = popen(command.c_str(), "r");
   if(!pipe)
      throw Exception("Unix_EntropySource: popen failed");
   while(!ferror(pipe) && !feof(pipe))
      {
      u32bit got = fread(output, 1, BUFFERSIZE, pipe);
      total_output += got;
      xor_buf(buffer, output, got);
      }
   pclose(pipe);
   return total_output;
   }

/*************************************************
* Add sources to the list                        *
*************************************************/
void Unix_EntropySource::add_sources(const Source srcs[], u32bit count)
   {
   sources.insert(sources.end(), srcs, srcs + count);
   std::sort(sources.begin(), sources.end(), Source_Cmp);
   }

}
